## $Id: browser.py 9926 2013-01-30 18:59:11Z henrik $
##
## Copyright (C) 2012 Uli Fouquet & Henrik Bettermann
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
##
from datetime import datetime
import httplib
import urllib
from xml.dom.minidom import parseString
import grok
from zope.component import getUtility
from zope.catalog.interfaces import ICatalog
from waeup.kofa.interfaces import IUniversity
from waeup.kofa.payments.interfaces import IPaymentWebservice
from waeup.kofa.browser.layout import KofaPage, UtilityView
from waeup.kofa.students.viewlets import ApprovePaymentActionButton as APABStudent
from waeup.kofa.applicants.viewlets import ApprovePaymentActionButton as APABApplicant
from waeup.aaue.interfaces import academic_sessions_vocab
from kofacustom.nigeria.interswitch.browser import (
    InterswitchActionButtonStudent,
    InterswitchRequestWebserviceActionButtonStudent,
    InterswitchActionButtonApplicant,
    InterswitchRequestWebserviceActionButtonApplicant)
from waeup.aaue.interfaces import MessageFactory as _
from waeup.aaue.students.interfaces import ICustomStudentOnlinePayment
from waeup.aaue.applicants.interfaces import ICustomApplicantOnlinePayment

#ACCEPTED_IP = ('195.219.3.181', '195.219.3.184')
ACCEPTED_IP = None


# Kofa's webservice

class KofaFeeRequest(grok.View):
    grok.context(IUniversity)
    grok.name('feerequest')
    grok.require('waeup.Public')

    def update(self, PAYEE_ID=None):
        if PAYEE_ID == None:
            self.output = '-1'
            return
        real_ip = self.request.get('HTTP_X_FORWARDED_FOR', None)
        # We can forego the logging once eTranzact payments run smoothly
        # and the accepted IP addresses are used.
        if real_ip:
            self.context.logger.info('KofaFeeRequest called: %s' % real_ip)
        if real_ip  and ACCEPTED_IP:
            if real_ip not in  ACCEPTED_IP:
                self.output = '-4'
                return

        # It seems eTranzact sends a POST request with an empty body but the URL
        # contains a query string. So it's actually a GET request pretended
        # to be a POST request. Although this does not comply with the
        # RFC 2616 HTTP guidelines we may try to fetch the id from the QUERY_STRING
        # value of the request.
        #if PAYEE_ID is None:
        #    try:
        #        PAYEE_ID = self.request['QUERY_STRING'].split('=')[1]
        #    except:
        #        self.output = '-2'
        #        return

        cat = getUtility(ICatalog, name='payments_catalog')
        results = list(cat.searchResults(p_id=(PAYEE_ID, PAYEE_ID)))
        if len(results) != 1:
            self.output = '-1'
            return
        try:
            owner = IPaymentWebservice(results[0])
            full_name = owner.display_fullname
            matric_no = owner.id
            faculty = owner.faculty
            department = owner.department
        except (TypeError, AttributeError):
            self.output = '-3'
            return
        amount = results[0].amount_auth
        payment_type = results[0].category
        programme_type = results[0].p_item
        academic_session = academic_sessions_vocab.getTerm(
            results[0].p_session).title
        status = results[0].p_state
        self.output = (
            'FULL_NAME=%s&' +
            'FACULTY=%s&' +
            'DEPARTMENT=%s&' +
            'RETURN_TYPE=%s&' +
            'PROGRAMME_TYPE=%s&' +
            'PAYMENT_TYPE=%s&' +
            'ACADEMIC_SESSION=%s&' +
            'MATRIC_NO=%s&' +
            'FEE_AMOUNT=%s&' +
            'TRANSACTION_STATUS=%s') % (full_name, faculty,
            department, PAYEE_ID, programme_type, payment_type,
            academic_session, matric_no, amount, status)
        return

    def render(self):
        return self.output


# Requerying eTranzact payments

TERMINAL_ID = '0330000046'
QUERY_URL =   'https://www.etranzact.net/Query/queryPayoutletTransaction.jsp'

# Test environment
#QUERY_URL =   'http://demo.etranzact.com:8080/WebConnect/queryPayoutletTransaction.jsp'
#TERMINAL_ID = '5009892289'

def query_etranzact(confirmation_number, payment):
    
    postdict = {}
    postdict['TERMINAL_ID'] = TERMINAL_ID
    #postdict['RESPONSE_URL'] = 'http://dummy'
    postdict['CONFIRMATION_NO'] = confirmation_number
    data = urllib.urlencode(postdict)
    payment.conf_number = confirmation_number
    try:
        f = urllib.urlopen(url=QUERY_URL, data=data)
        success = f.read()
        success = success.replace('\r\n','')
        if 'COL1' not in success:
            msg = _('Invalid or unsuccessful callback: ${a}',
                mapping = {'a': success})
            log = 'invalid callback for payment %s: %s' % (payment.p_id, success)
            payment.p_state = 'failed'
            return False, msg, log
        success = success.replace('%20',' ').split('&')
        # We expect at least two parameters
        if len(success) < 2:
            msg = _('Invalid callback: ${a}', mapping = {'a': success})
            log = 'invalid callback for payment %s: %s' % (payment.p_id, success)
            payment.p_state = 'failed'
            return False, msg, log
        try:
            success_dict = dict([tuple(i.split('=')) for i in success])
        except ValueError:
            msg = _('Invalid callback: ${a}', mapping = {'a': success})
            log = 'invalid callback for payment %s: %s' % (payment.p_id, success)
            payment.p_state = 'failed'
            return False, msg, log
    except IOError:
        msg = _('eTranzact IOError')
        log = 'eTranzact IOError'
        return False, msg, log
    payment.r_code = u'ET'
    payment.r_company = u'etranzact'
    payment.r_desc = u'%s' % success_dict.get('TRANS_DESCR')
    payment.r_amount_approved = float(success_dict.get('TRANS_AMOUNT',0.0))
    payment.r_card_num = None
    payment.r_pay_reference = u'%s' % success_dict.get('RECEIPT_NO')
    if payment.r_amount_approved != payment.amount_auth:
        msg = _('Wrong amount')
        log = 'wrong callback for payment %s: %s' % (payment.p_id, success)
        payment.p_state = 'failed'
        return False, msg, log
    #tcode = payment.p_id
    #tcode = tcode[len(tcode)-8:len(tcode)]
    col1 = success_dict.get('COL1')
    #col1 = col1[len(col1)-8:len(col1)]
    #if tcode != col1:
    if payment.p_id != col1:
        #msg = _('Wrong transaction code')
        msg = _('Wrong payment id')
        log = 'wrong callback for payment %s: %s' % (payment.p_id, success)
        payment.p_state = 'failed'
        return False, msg, log
    log = 'valid callback for payment %s: %s' % (payment.p_id, success)
    msg = _('Successful callback received')
    payment.p_state = 'paid'
    payment.payment_date = datetime.utcnow()
    return True, msg, log

class EtranzactEnterPinActionButtonApplicant(APABApplicant):
    grok.context(ICustomApplicantOnlinePayment)
    grok.require('waeup.payApplicant')
    grok.order(3)
    icon = 'actionicon_call.png'
    text = _('Query eTranzact History')
    target = 'enterpin'

class EtranzactEnterPinActionButtonStudent(APABStudent):
    grok.context(ICustomStudentOnlinePayment)
    grok.require('waeup.payStudent')
    grok.order(3)
    icon = 'actionicon_call.png'
    text = _('Query eTranzact History')
    target = 'enterpin'

class EtranzactEnterPinPageStudent(KofaPage):
    """
    """
    grok.context(ICustomStudentOnlinePayment)
    grok.name('enterpin')
    grok.template('enterpin')
    grok.require('waeup.payStudent')

    buttonname = _('Submit to eTranzact')
    label = _('Requery eTranzact History')
    action = 'query_history'

class EtranzactEnterPinPageApplicant(EtranzactEnterPinPageStudent):
    """
    """
    grok.require('waeup.payApplicant')
    grok.context(ICustomApplicantOnlinePayment)

class EtranzactQueryHistoryPageStudent(UtilityView, grok.View):
    """ Query history of eTranzact payments
    """
    grok.context(ICustomStudentOnlinePayment)
    grok.name('query_history')
    grok.require('waeup.payStudent')

    def update(self, confirmation_number=None):
        if self.context.p_state == 'paid':
            self.flash(_('This ticket has already been paid.'))
            return
        student = self.context.student
        success, msg, log = query_etranzact(confirmation_number,self.context)
        student.writeLogMessage(self, log)
        if not success:
            self.flash(msg)
            return
        success, msg, log = self.context.doAfterStudentPayment()
        if log is not None:
            student.writeLogMessage(self, log)
        self.flash(msg)
        return

    def render(self):
        self.redirect(self.url(self.context, '@@index'))
        return

class EtranzactQueryHistoryPageApplicant(UtilityView, grok.View):
    """ Query history of eTranzact payments
    """
    grok.context(ICustomApplicantOnlinePayment)
    grok.name('query_history')
    grok.require('waeup.payApplicant')

    def update(self, confirmation_number=None):
        ob_class = self.__implemented__.__name__
        if self.context.p_state == 'paid':
            self.flash(_('This ticket has already been paid.'))
            return
        applicant = self.context.__parent__
        success, msg, log = query_etranzact(confirmation_number,self.context)
        applicant.writeLogMessage(self, log)
        if not success:
            self.flash(msg)
            return
        success, msg, log = self.context.doAfterApplicantPayment()
        if log is not None:
            applicant.writeLogMessage(self, log)
        self.flash(msg)
        return

    def render(self):
        self.redirect(self.url(self.context, '@@index'))
        return

# Disable Interswitch viewlets. This could be avoided by defining the
# action button viewlets of kofacustom.nigeria.interswitch.browser in the
# context of INigeriaStudentOnlinePayment or INigeriaApplicantOnlinePayment
# respectively. But then all interswitch.browser modules have to be extended.

class InterswitchActionButtonStudent(InterswitchActionButtonStudent):

    @property
    def target_url(self):
        return ''

class InterswitchRequestWebserviceActionButtonStudent(
    InterswitchRequestWebserviceActionButtonStudent):

    @property
    def target_url(self):
        return ''

class InterswitchActionButtonApplicant(InterswitchActionButtonApplicant):

    @property
    def target_url(self):
        return ''

class InterswitchRequestWebserviceActionButtonApplicant(
    InterswitchRequestWebserviceActionButtonApplicant):

    @property
    def target_url(self):
        return ''